<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">
<html>
    <head>
        <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
        <title>Directory listing for {{ escape(path) }}</title>
        <link rel="stylesheet" href="/static/styles.css" />
    </head>
    <body>
        <div id="wrapper">
            <div id="breadcrumb">
                {% for (display, part_path) in breadcrumbs[:-1] %}
                <a href="{{ part_path }}">{{ display }}</a> /
                {% end %}
                <strong>{{ breadcrumbs[-1][0] }}</strong>
            </div>
            {% if files_hidden %}
            <div id="message">
                {% if hidden_shown %}
                    <a href=".">Hide hidden files</a>
                {% else %}
                    <a href="?hidden=1">Show hidden files</a>
                {% end %}
            </div>
            {% end %}
            <table>
                <tr>
                    <th class="dir-name">Name</th>
                    <th class="dir-size">Size</th>
                </tr>
                {% for (display, link, size) in dir_contents %}
                <tr>
                    <td class="dir-name"><a href="{{ link }}">{{ display }}</a></td>
                    <td class="dir-size">{% if size != None %}{{ size }}{% else %}&nbsp;{% end %}</td>
                </tr>
                {% end %}
            </table>
        </div>
    </body>
</html>
