((nil . ())
 (python-mode . ((indent-tabs-mode . nil)
                 (python-indent-offset . 4)))
 (sh-mode . ((indent-tabs-mode . nil)
             (sh-basic-offset . 4)))
 (c-mode . ((indent-tabs-mode . nil)
            (c-basic-offset . 4)
            (c-file-style . "BSD"))))
